package com.topcoder.ReviewTest;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.*;
import java.security.*;
import java.util.*;
import javax.swing.*;

public class PolyominoCoveringVis {

  // test parameters and constants
  private static int minN = 8, maxN = 50;
  private static int minNum = -9, maxNum = 9;
  private static int minTile = 2, maxTile = 7;
  final static int[] dr = { 0, 1, 0, -1 };
  final static int[] dc = { -1, 0, 1, 0 };
  final static int Penalty = 1000;
  final static int UncoveredCell = -1;

  // inputs
  int N;
  int[][] Grid;
  int[] Tiles;

  // score computation
  Tile T;
  Map<Integer, Tile> AllTiles; // ID -> tile
  long Score;

  // -----------------------------------------
  void generate(String seedStr) {
    try {
      // generate test case
      SecureRandom r1 = SecureRandom.getInstance("SHA1PRNG");
      long seed = Long.parseLong(seedStr);
      r1.setSeed(seed);

      // generate grid size
      N = r1.nextInt(maxN - minN + 1) + minN;

      if (seed == 1)
        N = minN;
      if (seed == 2)
        N = maxN;

      // generate grid of numbers
      Grid = new int[N][N];
      for (int i = 0; i < N; i++)
        for (int k = 0; k < N; k++)
          Grid[i][k] = r1.nextInt(maxNum - minNum + 1) + minNum;

      // generate tiles
      Tiles = new int[maxTile - minTile + 1];
      int cellsLeft = N * N;

      while (cellsLeft > 0) {
        int num = r1.nextInt(maxTile - minTile + 1) + minTile;
        if (cellsLeft - num < minTile && cellsLeft - num != 0)
          continue; // cannot use this tile

        Tiles[num - minTile]++;
        cellsLeft -= num;
      }

      if (Debug)
        printTest(seedStr);

    } catch (Exception e) {
      addFatalError("An exception occurred while generating test case.");
      e.printStackTrace();
    }
  }

  void printTest(String seed) {
    System.out.println("seed = " + seed);
    System.out.println("N = " + N);
    System.out.print("Tiles:");
    for (int i : Tiles)
      System.out.print(" " + i);
    System.out.println();
    System.out.println("Grid:");
    for (int i = 0; i < N; i++) {
      for (int k = 0; k < N; k++)
        System.out.print(Grid[i][k] + " ");
      System.out.println();
    }
  }

  // -----------------------------------------
  public double runTest(String seed) {
    try {
      generate(seed);

      if (proc != null) {
        // call the solution
        try {
          Solution = callSolution();
        } catch (Exception e) {
          addFatalError("Failed to get result from your solution.");
          return -1.0;
        }

        // check the return and score it
        if (Solution == null) {
          addFatalError("Your return contained invalid number of elements.");
          return -1.0;
        }
        if (Solution.length != N * N) {
          addFatalError("Your return did not contain " + (N * N) + " elements.");
          return -1.0;
        }

        Board = new int[N][N];
        for (int i = 0, cur = 0; i < N; i++)
          for (int k = 0; k < N; k++, cur++) {
            Board[i][k] = Solution[cur];

            if (!(Board[i][k] == UncoveredCell || Board[i][k] > 0)) {
              addFatalError("Cell (" + i + "," + k + ") has illegal ID " + Board[i][k]);
              return -1.0;
            }
          }

        // compute score
        boolean[][] seen = new boolean[N][N];
        int[] numUsed = new int[maxTile + 1];
        AllTiles = new HashMap<Integer, Tile>();
        Score = 0;

        for (int r = 0; r < N; r++) {
          for (int c = 0; c < N; c++) {
            if (seen[r][c])
              continue;

            int id = Board[r][c];

            // cell not covered
            if (id == UncoveredCell) {
              Score -= Penalty;
            } else {
              T = new Tile();
              go(r, c, id, seen);

              if (T.size < minTile) {
                addFatalError("Tile " + id + " is too small");
                return -1.0;
              }
              if (T.size > maxTile) {
                addFatalError("Tile " + id + " is too big");
                return -1.0;
              }
              if (AllTiles.containsKey(id)) {
                addFatalError("Tile " + id + " has already been used");
                return -1.0;
              }

              AllTiles.put(id, T);
              Score += T.score;

              numUsed[T.size]++;
              if (numUsed[T.size] > Tiles[T.size - minTile]) {
                addFatalError("You have used too many tiles of size " + T.size);
                return -1.0;
              }
            }
          }
        }

        if (vis) {
          jf.setSize(N * SZ + 150, N * SZ + 50);
          jf.setVisible(true);
          draw();
        }
      }

      return this.score = Score;
    } catch (Exception e) {
      addFatalError("An exception occurred while trying to get your program's results.");
      e.printStackTrace();
      return -1.0;
    }
  }

  void go(int r, int c, int id, boolean[][] seen) {
    if (r < 0 || r >= N || c < 0 || c >= N || seen[r][c] || Board[r][c] != id)
      return;

    seen[r][c] = true;
    T.loc.add(new Loc(r, c));
    T.score *= Grid[r][c];
    T.size++;

    for (int m = 0; m < dr.length; m++)
      go(r + dr[m], c + dc[m], id, seen);
  }

  // ------------- visualization part ------------
  JFrame jf;
  Vis v;
  static String exec;
  static boolean vis;
  static Process proc;
  InputStream is;
  OutputStream os;
  BufferedReader br;
  static int SZ;
  static boolean Debug;
  static int[] Solution;
  static int[][] Board;

  // -----------------------------------------
  private int[] callSolution() throws IOException, NumberFormatException {
    if (exec == null)
      return null;

    String s = N + "\n";
    os.write(s.getBytes());
    for (int i = 0; i < N; i++)
      for (int k = 0; k < N; k++) {
        s = Grid[i][k] + "\n";
        os.write(s.getBytes());
      }

    for (int i : Tiles) {
      s = i + "\n";
      os.write(s.getBytes());
    }
    os.flush();

    long startTime = System.currentTimeMillis();

    // and get the return value
    int[] grid = new int[Integer.parseInt(br.readLine())];
    for (int i = 0; i < grid.length; i++)
      grid[i] = Integer.parseInt(br.readLine());

    runTime += System.currentTimeMillis() - startTime;

    return grid;
  }

  // -----------------------------------------
  void draw() {
    if (!vis)
      return;
    v.repaint();
  }

  // -----------------------------------------
  public class Vis extends JPanel implements MouseListener, WindowListener {
    public void paint(Graphics g) {

      Color[] colors = { null, null, Color.GRAY, Color.GREEN, Color.CYAN, Color.YELLOW, Color.PINK, Color.MAGENTA };

      int startX = 10;
      int startY = 10;

      // background
      g.setColor(new Color(0xDDDDDD));
      g.fillRect(0, 0, N * SZ + 150, N * SZ + 40);
      g.setColor(Color.WHITE);
      g.fillRect(startX, startY, N * SZ, N * SZ);

      // paint thin lines between cells
      g.setColor(Color.BLACK);
      for (int i = 0; i <= N; i++)
        g.drawLine(startX, startY + i * SZ, startX + N * SZ, startY + i * SZ);
      for (int i = 0; i <= N; i++)
        g.drawLine(startX + i * SZ, startY, startX + i * SZ, startY + N * SZ);

      for (int r = 0; r < N; r++)
        for (int c = 0; c < N; c++) {
          // paint tiles
          if (Board[r][c] != UncoveredCell) {
            Tile t = AllTiles.get(Board[r][c]);
            g.setColor(colors[t.size]);
            g.fillRect(startX + c * SZ + 1, startY + r * SZ + 1, SZ - 1, SZ - 1);
          }

          // paint numbers
          g.setFont(new Font("Arial", Font.PLAIN, 14));
          g.setColor(Color.BLACK);
          g.drawString("" + Grid[r][c], startX + c * SZ + SZ / 2 - 5, startY + r * SZ + SZ / 2 + 5);
        }

      // display score
      g.setColor(Color.BLACK);
      g.setFont(new Font("Arial", Font.BOLD, 14));
      g.drawString("SCORE", SZ * N + 25, 30);
      g.drawString(String.format("%d", Score), SZ * N + 25, 50);

      // draw thick lines for tile borders
      Graphics2D g2 = (Graphics2D) g.create();
      g2.setColor(Color.BLACK);
      g2.setStroke(new BasicStroke(4));

      for (int r = 0; r < N; r++)
        for (int c = 0; c < N; c++) {
          int r2 = r;
          int c2 = c + 1;
          if (c2 < N && Board[r][c] != Board[r2][c2])
            g2.drawLine(startX + (c + 1) * SZ, startY + r * SZ, startX + (c + 1) * SZ, startY + (r + 1) * SZ);

          r2 = r + 1;
          c2 = c;
          if (r2 < N && Board[r][c] != Board[r2][c2])
            g2.drawLine(startX + c * SZ, startY + (r + 1) * SZ, startX + (c + 1) * SZ, startY + (r + 1) * SZ);
        }

      // paint thick border around grid
      g2.drawRect(startX, startY, N * SZ, N * SZ);
      g2.dispose();

    }

    // -------------------------------------
    public Vis() {
      jf.addWindowListener(this);
    }

    // -------------------------------------
    // WindowListener
    public void windowClosing(WindowEvent e) {
      if (proc != null)
        try {
          proc.destroy();
        } catch (Exception ex) {
          ex.printStackTrace();
        }
      System.exit(0);
    }

    public void windowActivated(WindowEvent e) {
    }

    public void windowDeactivated(WindowEvent e) {
    }

    public void windowOpened(WindowEvent e) {
    }

    public void windowClosed(WindowEvent e) {
    }

    public void windowIconified(WindowEvent e) {
    }

    public void windowDeiconified(WindowEvent e) {
    }

    // -------------------------------------
    // MouseListener
    public void mouseClicked(MouseEvent e) {
    }

    public void mousePressed(MouseEvent e) {
    }

    public void mouseReleased(MouseEvent e) {
    }

    public void mouseEntered(MouseEvent e) {
    }

    public void mouseExited(MouseEvent e) {
    }
  }

  // -----------------------------------------
  public PolyominoCoveringVis(String seed) {
    try {
      if (vis) {
        jf = new JFrame();
        v = new Vis();
        jf.getContentPane().add(v);
      }
      if (exec != null) {
        try {
          Runtime rt = Runtime.getRuntime();
          proc = rt.exec(exec);
          os = proc.getOutputStream();
          is = proc.getInputStream();
          br = new BufferedReader(new InputStreamReader(is));
          output = new ErrorReader(proc.getErrorStream());
          output.start();
        } catch (Exception e) {
          e.printStackTrace();
        }
      }
      System.out.println("Score = " + runTest(seed));
      if (proc != null)
        try {
          proc.destroy();
        } catch (Exception e) {
          e.printStackTrace();
        }
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  class Loc {
    int r;
    int c;

    public Loc(int r2, int c2) {
      r = r2;
      c = c2;
    }
  }

  class Tile {
    java.util.List loc;
    int score;
    int size;

    public Tile() {
      loc = new ArrayList<Loc>();
      score = 1;
      size = 0;
    }
  }

  public static long runTime = 0;
  public double score = -1.0;
  public static ErrorReader output;

  // -----------------------------------------
  public static double main(String[] args) {
    String seed = "1";
    vis = true;
    SZ = 20;
    Debug = false;
    runTime = 0;
    for (int i = 0; i < args.length; i++) {
      if (args[i].equals("-seed"))
        seed = args[++i];
      if (args[i].equals("-exec"))
        exec = args[++i];
      if (args[i].equals("-novis"))
        vis = false;
      if (args[i].equals("-size"))
        SZ = Integer.parseInt(args[++i]);
      if (args[i].equals("-debug"))
        Debug = true;
    }
    if (seed.equals("1"))
      SZ = 50;

    PolyominoCoveringVis f = new PolyominoCoveringVis(seed);
    return f.score;
  }

  // -----------------------------------------
  void addFatalError(String message) {
    System.out.println(message);
  }
}

class ErrorReader extends Thread {
  InputStream error;
  StringBuilder sb = new StringBuilder();

  public ErrorReader(InputStream is) {
    error = is;
  }

  public void run() {
    try {
      byte[] ch = new byte[50000];
      int read;
      while ((read = error.read(ch)) > 0) {
        String s = new String(ch, 0, read);
        System.out.print(s);
        System.out.flush();
      }
    } catch (Exception e) {
    }
  }

  public String getOutput() {
    return sb.toString();
  }
}
